@@ -23,7 +23,7 @@ import (
 	"context"
 	"encoding/json"
 	"fmt"
-	"net/url"
+	neturl "net/url"
 	"reflect"
 
 	admissionv1 "k8s.io/api/admission/v1"
@@ -36,7 +36,7 @@ import (
 	"k8s.io/client-go/kubernetes"
 	"k8s.io/klog/v2"
 
-	cdiv1 "kubevirt.io/containerized-data-importer/pkg/apis/core/v1beta1"
+	cdiv1 "kubevirt.io/containerized-data-importer-api/pkg/apis/core/v1beta1"
 	cdiclient "kubevirt.io/containerized-data-importer/pkg/client/clientset/versioned"
 	"kubevirt.io/containerized-data-importer/pkg/controller"
 )
@@ -50,7 +50,7 @@ func validateSourceURL(sourceURL string) string {
 	if sourceURL == "" {
 		return "source URL is empty"
 	}
-	url, err := url.ParseRequestURI(sourceURL)
+	url, err := neturl.ParseRequestURI(sourceURL)
 	if err != nil {
 		return fmt.Sprintf("Invalid source URL: %s", sourceURL)
 	}
@@ -60,12 +60,12 @@ func validateSourceURL(sourceURL string) string {
 	return ""
 }
 
-func validateDataVolumeName(name string) []metav1.StatusCause {
+func validateNameLength(name string, maxLen int) []metav1.StatusCause {
 	var causes []metav1.StatusCause
-	if len(name) > kvalidation.DNS1123SubdomainMaxLength {
+	if len(name) > maxLen {
 		causes = append(causes, metav1.StatusCause{
 			Type:    metav1.CauseTypeFieldValueInvalid,
-			Message: fmt.Sprintf("Name of data volume cannot be more than %d characters", kvalidation.DNS1123SubdomainMaxLength),
+			Message: fmt.Sprintf("Name cannot be longer than %d characters", maxLen),
 			Field:   "",
 		})
 	}
@@ -240,14 +240,22 @@ func (wh *dataVolumeValidatingWebhook) validateDataVolumeSpec(request *admission
 		return causes
 	}
 
-	if spec.Source.Registry != nil && spec.ContentType != "" && string(spec.ContentType) != string(cdiv1.DataVolumeKubeVirt) {
-		sourceType = field.Child("contentType").String()
-		causes = append(causes, metav1.StatusCause{
-			Type:    metav1.CauseTypeFieldValueInvalid,
-			Message: fmt.Sprintf("ContentType must be " + string(cdiv1.DataVolumeKubeVirt) + " when Source is Registry"),
-			Field:   sourceType,
-		})
-		return causes
+	if spec.Source.Registry != nil {
+		if spec.ContentType != "" && string(spec.ContentType) != string(cdiv1.DataVolumeKubeVirt) {
+			sourceType = field.Child("contentType").String()
+			causes = append(causes, metav1.StatusCause{
+				Type:    metav1.CauseTypeFieldValueInvalid,
+				Message: fmt.Sprintf("ContentType must be %s when Source is Registry", cdiv1.DataVolumeKubeVirt),
+				Field:   sourceType,
+			})
+			return causes
+		}
+
+		causes := validateDataVolumeSourceRegistry(spec.Source.Registry, field)
+		if len(causes) > 0 {
+			return causes
+		}
+
 	}
 
 	if spec.Source.Imageio != nil {
@@ -292,6 +300,69 @@ func (wh *dataVolumeValidatingWebhook) validateDataVolumeSpec(request *admission
 	return causes
 }
 
+func validateDataVolumeSourceRegistry(sourceRegistry *cdiv1.DataVolumeSourceRegistry, field *k8sfield.Path) []metav1.StatusCause {
+	var causes []metav1.StatusCause
+	sourceURL := sourceRegistry.URL
+	sourceIS := sourceRegistry.ImageStream
+	if (sourceURL == nil && sourceIS == nil) || (sourceURL != nil && sourceIS != nil) {
+		causes = append(causes, metav1.StatusCause{
+			Type:    metav1.CauseTypeFieldValueInvalid,
+			Message: fmt.Sprintf("Source registry should have either URL or ImageStream"),
+			Field:   field.Child("source", "Registry").String(),
+		})
+		return causes
+	}
+	if sourceURL != nil {
+		url, err := neturl.Parse(*sourceURL)
+		if err != nil {
+			causes = append(causes, metav1.StatusCause{
+				Type:    metav1.CauseTypeFieldValueInvalid,
+				Message: fmt.Sprintf("Illegal registry source URL %s", *sourceURL),
+				Field:   field.Child("source", "Registry", "URL").String(),
+			})
+			return causes
+		}
+		scheme := url.Scheme
+		if scheme != cdiv1.RegistrySchemeDocker && scheme != cdiv1.RegistrySchemeOci {
+			causes = append(causes, metav1.StatusCause{
+				Type:    metav1.CauseTypeFieldValueInvalid,
+				Message: fmt.Sprintf("Illegal registry source URL scheme %s", url),
+				Field:   field.Child("source", "Registry", "URL").String(),
+			})
+			return causes
+		}
+	}
+	importMethod := sourceRegistry.PullMethod
+	if importMethod != nil && *importMethod != cdiv1.RegistryPullPod && *importMethod != cdiv1.RegistryPullNode {
+		causes = append(causes, metav1.StatusCause{
+			Type:    metav1.CauseTypeFieldValueInvalid,
+			Message: fmt.Sprintf("ImportMethod %s is neither %s, %s or \"\"", *importMethod, cdiv1.RegistryPullPod, cdiv1.RegistryPullNode),
+			Field:   field.Child("source", "Registry", "importMethod").String(),
+		})
+		return causes
+	}
+
+	if sourceIS != nil && *sourceIS == "" {
+		causes = append(causes, metav1.StatusCause{
+			Type:    metav1.CauseTypeFieldValueInvalid,
+			Message: fmt.Sprintf("Source registry ImageStream is not valid"),
+			Field:   field.Child("source", "Registry", "importMethod").String(),
+		})
+		return causes
+	}
+
+	if sourceIS != nil && (importMethod == nil || *importMethod != cdiv1.RegistryPullNode) {
+		causes = append(causes, metav1.StatusCause{
+			Type:    metav1.CauseTypeFieldValueInvalid,
+			Message: fmt.Sprintf("Source registry ImageStream is supported only with node pull import method"),
+			Field:   field.Child("source", "Registry", "importMethod").String(),
+		})
+		return causes
+	}
+
+	return causes
+}
+
 func (wh *dataVolumeValidatingWebhook) validateSourceRef(request *admissionv1.AdmissionRequest, spec *cdiv1.DataVolumeSpec, field *k8sfield.Path, namespace *string) *metav1.StatusCause {
 	if spec.SourceRef.Kind == "" {
 		return &metav1.StatusCause{
@@ -429,7 +500,8 @@ func (wh *dataVolumeValidatingWebhook) Admit(ar admissionv1.AdmissionReview) *ad
 
 		// Always admit checkpoint updates for multi-stage migrations.
 		multiStageAdmitted := false
-		isMultiStage := dv.Spec.Source != nil && dv.Spec.Source.VDDK != nil && len(dv.Spec.Checkpoints) > 0
+		isMultiStage := dv.Spec.Source != nil && len(dv.Spec.Checkpoints) > 0 &&
+			(dv.Spec.Source.VDDK != nil || dv.Spec.Source.Imageio != nil)
 		if isMultiStage {
 			oldSpec := oldDV.Spec.DeepCopy()
 			oldSpec.FinalCheckpoint = false
@@ -454,7 +526,7 @@ func (wh *dataVolumeValidatingWebhook) Admit(ar admissionv1.AdmissionReview) *ad
 		}
 	}
 
-	causes := validateDataVolumeName(dv.Name)
+	causes := validateNameLength(dv.Name, kvalidation.DNS1123SubdomainMaxLength)
 	if len(causes) > 0 {
 		klog.Infof("rejected DataVolume admission")
 		return toRejectedAdmissionResponse(causes)
