@@ -3,34 +3,37 @@ package controller
 import (
 	"context"
 	"crypto/rsa"
+	"encoding/json"
 	"fmt"
+	"regexp"
+	"strconv"
 	"strings"
 
-	featuregates "kubevirt.io/containerized-data-importer/pkg/feature-gates"
-	"kubevirt.io/containerized-data-importer/pkg/token"
-
 	"github.com/go-logr/logr"
 	snapshotv1 "github.com/kubernetes-csi/external-snapshotter/v2/pkg/apis/volumesnapshot/v1beta1"
+	ocpconfigv1 "github.com/openshift/api/config/v1"
 	"github.com/pkg/errors"
-	corev1 "k8s.io/api/core/v1"
 	v1 "k8s.io/api/core/v1"
 	storagev1 "k8s.io/api/storage/v1"
 	extclientset "k8s.io/apiextensions-apiserver/pkg/client/clientset/clientset"
 	k8serrors "k8s.io/apimachinery/pkg/api/errors"
+	"k8s.io/apimachinery/pkg/api/resource"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/types"
 	"k8s.io/apimachinery/pkg/util/sets"
+	"k8s.io/client-go/tools/record"
 	"k8s.io/klog/v2"
 	"sigs.k8s.io/controller-runtime/pkg/client"
 
-	cdiv1 "kubevirt.io/containerized-data-importer/pkg/apis/core/v1beta1"
-	cdiv1utils "kubevirt.io/containerized-data-importer/pkg/apis/core/v1beta1/utils"
+	cdiv1 "kubevirt.io/containerized-data-importer-api/pkg/apis/core/v1beta1"
+	cdiv1utils "kubevirt.io/containerized-data-importer-api/pkg/apis/core/v1beta1/utils"
 	"kubevirt.io/containerized-data-importer/pkg/common"
+	featuregates "kubevirt.io/containerized-data-importer/pkg/feature-gates"
+	"kubevirt.io/containerized-data-importer/pkg/token"
+	"kubevirt.io/containerized-data-importer/pkg/util"
 	"kubevirt.io/containerized-data-importer/pkg/util/cert"
 	"kubevirt.io/containerized-data-importer/pkg/util/naming"
 	sdkapi "kubevirt.io/controller-lifecycle-operator-sdk/pkg/sdk/api"
-
-	ocpconfigv1 "github.com/openshift/api/config/v1"
 )
 
 const (
@@ -65,6 +68,8 @@ const (
 	AnnPrePopulated = AnnAPIGroup + "/storage.prePopulated"
 	// AnnPriorityClassName is PVC annotation to indicate the priority class name for importer, cloner and uploader pod
 	AnnPriorityClassName = AnnAPIGroup + "/storage.pod.priorityclassname"
+	// AnnPodRetainAfterCompletion is PVC annotation for retaining transfer pods after completion)
+	AnnPodRetainAfterCompletion = AnnAPIGroup + "/storage.pod.retainAfterCompletion"
 
 	// AnnPreviousCheckpoint provides a const to indicate the previous snapshot for a multistage import
 	AnnPreviousCheckpoint = AnnAPIGroup + "/storage.checkpoint.previous"
@@ -109,6 +114,8 @@ const (
 	AnnVddkVersion = AnnAPIGroup + "/storage.pod.vddk.version"
 	// AnnVddkHostConnection shows the last ESX host that serviced a DV's importer pod
 	AnnVddkHostConnection = AnnAPIGroup + "/storage.pod.vddk.host"
+	// AnnVddkInitImageURL saves a per-DV VDDK image URL on the PVC
+	AnnVddkInitImageURL = AnnAPIGroup + "/storage.pod.vddk.initimageurl"
 
 	// PodRunningReason is const that defines the pod was started as a reason
 	podRunningReason = "Pod is running"
@@ -132,6 +139,9 @@ const (
 	ClusterWideProxyConfigMapNameSpace = "openshift-config"
 	// ClusterWideProxyConfigMapKey is the OpenShift Cluster Wide Proxy ConfigMap key name for CA certificates.
 	ClusterWideProxyConfigMapKey = "ca-bundle.crt"
+
+	// SecretExtraHeadersVolumeName is the format string that specifies where extra HTTP header secrets will be mounted
+	SecretExtraHeadersVolumeName = "cdi-secret-extra-headers-vol-%d"
 )
 
 const (
@@ -145,6 +155,10 @@ const (
 	AnnPodSidecarInjectionDefault = "false"
 )
 
+var (
+	vddkInfoMatch = regexp.MustCompile(`((.*; )|^)VDDK: (?P<info>{.*})`)
+)
+
 func isCrossNamespaceClone(dv *cdiv1.DataVolume) bool {
 	if dv.Spec.Source.PVC == nil {
 		return false
@@ -192,31 +206,6 @@ func isBound(pvc *v1.PersistentVolumeClaim, log logr.Logger) bool {
 	return true
 }
 
-func isPvcUsedByAnyPod(c client.Client, pvc *v1.PersistentVolumeClaim, log logr.Logger) (bool, error) {
-	pods := &v1.PodList{}
-	if err := c.List(context.TODO(), pods, &client.ListOptions{Namespace: pvc.Namespace}); err != nil {
-		return false, errors.Wrap(err, "error listing pods")
-	}
-
-	for _, pod := range pods.Items {
-		if isPvcUsedByPod(pod, pvc.Name) {
-			return true, nil
-		}
-	}
-
-	return false, nil
-}
-
-func isPvcUsedByPod(pod v1.Pod, pvcName string) bool {
-	for _, volume := range pod.Spec.Volumes {
-		if volume.VolumeSource.PersistentVolumeClaim != nil &&
-			volume.PersistentVolumeClaim.ClaimName == pvcName {
-			return true
-		}
-	}
-	return false
-}
-
 func getRequestedImageSize(pvc *v1.PersistentVolumeClaim) (string, error) {
 	pvcSize, found := pvc.Spec.Resources.Requests[v1.ResourceStorage]
 	if !found {
@@ -225,6 +214,11 @@ func getRequestedImageSize(pvc *v1.PersistentVolumeClaim) (string, error) {
 	return pvcSize.String(), nil
 }
 
+// GetVolumeMode returns the volumeMode from PVC handling default empty value
+func GetVolumeMode(pvc *v1.PersistentVolumeClaim) v1.PersistentVolumeMode {
+	return getVolumeMode(pvc)
+}
+
 // returns the volumeMode from PVC handling default empty value
 func getVolumeMode(pvc *v1.PersistentVolumeClaim) v1.PersistentVolumeMode {
 	return resolveVolumeMode(pvc.Spec.VolumeMode)
@@ -256,7 +250,7 @@ func newScratchPersistentVolumeClaimSpec(pvc *v1.PersistentVolumeClaim, pod *v1.
 		"app": "containerized-data-importer",
 	}
 
-	annotations := make(map[string]string, 0)
+	annotations := make(map[string]string)
 	// Copy kubevirt.io annotations, but NOT the CDI annotations as those will trigger another import/upload/clone on the scratchspace
 	// pvc.
 	if len(pvc.GetAnnotations()) > 0 {
@@ -288,9 +282,13 @@ func newScratchPersistentVolumeClaimSpec(pvc *v1.PersistentVolumeClaim, pod *v1.
 }
 
 // CreateScratchPersistentVolumeClaim creates and returns a pointer to a scratch PVC which is created based on the passed-in pvc and storage class name.
-func CreateScratchPersistentVolumeClaim(client client.Client, pvc *v1.PersistentVolumeClaim, pod *v1.Pod, name, storageClassName string) (*v1.PersistentVolumeClaim, error) {
+func CreateScratchPersistentVolumeClaim(client client.Client, pvc *v1.PersistentVolumeClaim, pod *v1.Pod, name, storageClassName string, installerLabels map[string]string, recorder record.EventRecorder) (*v1.PersistentVolumeClaim, error) {
 	scratchPvcSpec := newScratchPersistentVolumeClaimSpec(pvc, pod, name, storageClassName)
+	util.SetRecommendedLabels(scratchPvcSpec, installerLabels, "cdi-controller")
 	if err := client.Create(context.TODO(), scratchPvcSpec); err != nil {
+		if errQuotaExceeded(err) {
+			recorder.Event(pvc, v1.EventTypeWarning, ErrExceededQuota, err.Error())
+		}
 		if !k8serrors.IsAlreadyExists(err) {
 			return nil, errors.Wrap(err, "scratch PVC API create errored")
 		}
@@ -337,7 +335,6 @@ func GetDefaultStorageClass(client client.Client) (*storagev1.StorageClass, erro
 
 // GetFilesystemOverhead determines the filesystem overhead defined in CDIConfig for this PVC's volumeMode and storageClass.
 func GetFilesystemOverhead(client client.Client, pvc *v1.PersistentVolumeClaim) (cdiv1.Percent, error) {
-	klog.V(1).Info("GetFilesystemOverhead with PVC", pvc)
 	if getVolumeMode(pvc) != v1.PersistentVolumeFilesystem {
 		return "0", nil
 	}
@@ -546,28 +543,75 @@ func podSucceededFromPVC(pvc *v1.PersistentVolumeClaim) bool {
 	return (podPhaseFromPVC(pvc) == v1.PodSucceeded)
 }
 
-func setConditionFromPodWithPrefix(anno map[string]string, prefix string, pod *v1.Pod) {
-	if pod.Status.ContainerStatuses != nil {
-		if pod.Status.ContainerStatuses[0].State.Running != nil {
-			anno[prefix] = "true"
-			anno[prefix+".message"] = ""
-			anno[prefix+".reason"] = podRunningReason
-		} else {
-			anno[AnnRunningCondition] = "false"
-			if pod.Status.ContainerStatuses[0].State.Waiting != nil {
-				anno[prefix+".message"] = pod.Status.ContainerStatuses[0].State.Waiting.Message
-				anno[prefix+".reason"] = pod.Status.ContainerStatuses[0].State.Waiting.Reason
-			} else if pod.Status.ContainerStatuses[0].State.Terminated != nil {
-				anno[prefix+".message"] = pod.Status.ContainerStatuses[0].State.Terminated.Message
-				anno[prefix+".reason"] = pod.Status.ContainerStatuses[0].State.Terminated.Reason
-				if strings.Contains(pod.Status.ContainerStatuses[0].State.Terminated.Message, common.PreallocationApplied) {
-					anno[AnnPreallocationApplied] = "true"
-				}
+func setAnnotationsFromPodWithPrefix(anno map[string]string, pod *v1.Pod, prefix string) {
+	if pod == nil || pod.Status.ContainerStatuses == nil {
+		return
+	}
+	annPodRestarts, _ := strconv.Atoi(anno[AnnPodRestarts])
+	podRestarts := int(pod.Status.ContainerStatuses[0].RestartCount)
+	if podRestarts >= annPodRestarts {
+		anno[AnnPodRestarts] = strconv.Itoa(podRestarts)
+	}
+	setVddkAnnotations(anno, pod)
+	containerState := pod.Status.ContainerStatuses[0].State
+	if containerState.Running != nil {
+		anno[prefix] = "true"
+		anno[prefix+".message"] = ""
+		anno[prefix+".reason"] = podRunningReason
+	} else {
+		anno[AnnRunningCondition] = "false"
+		if containerState.Waiting != nil && containerState.Waiting.Reason != "CrashLoopBackOff" {
+			anno[prefix+".message"] = simplifyKnownMessage(containerState.Waiting.Message)
+			anno[prefix+".reason"] = containerState.Waiting.Reason
+		} else if containerState.Terminated != nil {
+			anno[prefix+".message"] = simplifyKnownMessage(containerState.Terminated.Message)
+			anno[prefix+".reason"] = containerState.Terminated.Reason
+			if strings.Contains(containerState.Terminated.Message, common.PreallocationApplied) {
+				anno[AnnPreallocationApplied] = "true"
 			}
 		}
 	}
 }
 
+func simplifyKnownMessage(msg string) string {
+	if strings.Contains(msg, "is larger than the reported available") ||
+		strings.Contains(msg, "no space left on device") ||
+		strings.Contains(msg, "file largest block is bigger than maxblock") {
+		return "DataVolume too small to contain image"
+	}
+
+	return msg
+}
+
+func setVddkAnnotations(anno map[string]string, pod *v1.Pod) {
+	if pod.Status.ContainerStatuses[0].State.Terminated == nil {
+		return
+	}
+	terminationMessage := pod.Status.ContainerStatuses[0].State.Terminated.Message
+	klog.V(1).Info("Saving VDDK annotations from pod status message: ", "message", terminationMessage)
+
+	var terminationInfo string
+	matches := vddkInfoMatch.FindAllStringSubmatch(terminationMessage, -1)
+	for index, matchName := range vddkInfoMatch.SubexpNames() {
+		if matchName == "info" && len(matches) > 0 {
+			terminationInfo = matches[0][index]
+			break
+		}
+	}
+
+	var vddkInfo util.VddkInfo
+	err := json.Unmarshal([]byte(terminationInfo), &vddkInfo)
+	if err != nil {
+		return
+	}
+	if vddkInfo.Host != "" {
+		anno[AnnVddkHostConnection] = vddkInfo.Host
+	}
+	if vddkInfo.Version != "" {
+		anno[AnnVddkVersion] = vddkInfo.Version
+	}
+}
+
 func setBoundConditionFromPVC(anno map[string]string, prefix string, pvc *v1.PersistentVolumeClaim) {
 	switch pvc.Status.Phase {
 	case v1.ClaimBound:
@@ -663,7 +707,7 @@ func GetActiveCDI(c client.Client) (*cdiv1.CDI, error) {
 	}
 
 	if len(activeResources) > 1 {
-		return nil, fmt.Errorf("Number of active CDI CRs > 1")
+		return nil, fmt.Errorf("number of active CDI CRs > 1")
 	}
 
 	return &activeResources[0], nil
@@ -765,6 +809,10 @@ func getPriorityClass(pvc *v1.PersistentVolumeClaim) string {
 	return anno[AnnPriorityClassName]
 }
 
+func shouldDeletePod(pvc *v1.PersistentVolumeClaim) bool {
+	return pvc.GetAnnotations()[AnnPodRetainAfterCompletion] != "true" || pvc.GetAnnotations()[AnnRequiresScratch] == "true" || pvc.DeletionTimestamp != nil
+}
+
 // AddFinalizer adds a finalizer to a resource
 func AddFinalizer(obj metav1.Object, name string) {
 	if HasFinalizer(obj, name) {
@@ -800,22 +848,17 @@ func HasFinalizer(object metav1.Object, value string) bool {
 	return false
 }
 
-func validateCloneTokenPVC(validator token.Validator, source, target *corev1.PersistentVolumeClaim) error {
+func validateCloneTokenPVC(t string, v token.Validator, source, target *v1.PersistentVolumeClaim) error {
 	if source.Namespace == target.Namespace {
 		return nil
 	}
 
-	tok, ok := target.Annotations[AnnCloneToken]
-	if !ok {
-		return errors.New("clone token missing")
-	}
-
-	tokenData, err := validator.Validate(tok)
+	tokenData, err := v.Validate(t)
 	if err != nil {
 		return errors.Wrap(err, "error verifying token")
 	}
 
-	return validateTokenData(tokenData, source.Namespace, source.Name, target.Namespace, target.Name)
+	return validateTokenData(tokenData, source.Namespace, source.Name, target.Namespace, target.Name, string(target.UID))
 }
 
 func validateCloneTokenDV(validator token.Validator, dv *cdiv1.DataVolume) error {
@@ -833,16 +876,18 @@ func validateCloneTokenDV(validator token.Validator, dv *cdiv1.DataVolume) error
 		return errors.Wrap(err, "error verifying token")
 	}
 
-	return validateTokenData(tokenData, dv.Spec.Source.PVC.Namespace, dv.Spec.Source.PVC.Name, dv.Namespace, dv.Name)
+	return validateTokenData(tokenData, dv.Spec.Source.PVC.Namespace, dv.Spec.Source.PVC.Name, dv.Namespace, dv.Name, "")
 }
 
-func validateTokenData(tokenData *token.Payload, srcNamespace, srcName, targetNamespace, targetName string) error {
+func validateTokenData(tokenData *token.Payload, srcNamespace, srcName, targetNamespace, targetName, targetUID string) error {
+	uid := tokenData.Params["uid"]
 	if tokenData.Operation != token.OperationClone ||
 		tokenData.Name != srcName ||
 		tokenData.Namespace != srcNamespace ||
 		tokenData.Resource.Resource != "persistentvolumeclaims" ||
 		tokenData.Params["targetNamespace"] != targetNamespace ||
-		tokenData.Params["targetName"] != targetName {
+		tokenData.Params["targetName"] != targetName ||
+		(uid != "" && uid != targetUID) {
 		return errors.New("invalid token")
 	}
 
@@ -855,3 +900,211 @@ func addAnnotation(obj metav1.Object, key, value string) {
 	}
 	obj.GetAnnotations()[key] = value
 }
+
+// RenderPvcSpec creates a new PVC Spec based on either the dv.spec.pvc or dv.spec.storage section
+func RenderPvcSpec(client client.Client, recorder record.EventRecorder, log logr.Logger, dv *cdiv1.DataVolume) (*v1.PersistentVolumeClaimSpec, error) {
+	if dv.Spec.PVC != nil {
+		return dv.Spec.PVC.DeepCopy(), nil
+	}
+
+	if dv.Spec.Storage != nil {
+		return pvcFromStorage(client, recorder, log, dv)
+	}
+
+	return nil, errors.Errorf("datavolume one of {pvc, storage} field is required")
+}
+
+func pvcFromStorage(client client.Client, recorder record.EventRecorder, log logr.Logger, dv *cdiv1.DataVolume) (*v1.PersistentVolumeClaimSpec, error) {
+	storage := dv.Spec.Storage
+	pvcSpec := copyStorageAsPvc(log, storage)
+
+	if dv.Spec.ContentType == cdiv1.DataVolumeArchive {
+		if pvcSpec.VolumeMode != nil && *pvcSpec.VolumeMode == v1.PersistentVolumeBlock {
+			log.V(1).Info("DataVolume with ContentType Archive cannot have block volumeMode", "namespace", dv.Namespace, "name", dv.Name)
+			recorder.Eventf(dv, v1.EventTypeWarning, ErrClaimNotValid, "DataVolume with ContentType Archive cannot have block volumeMode")
+			return nil, errors.Errorf("DataVolume with ContentType Archive cannot have block volumeMode")
+		}
+		volumeMode := v1.PersistentVolumeFilesystem
+		pvcSpec.VolumeMode = &volumeMode
+	}
+
+	storageClass, err := GetStorageClassByName(client, storage.StorageClassName)
+	if err != nil {
+		return nil, err
+	}
+
+	if storageClass == nil {
+		// Not even default storageClass on the cluster, cannot apply the defaults, verify spec is ok
+		if len(pvcSpec.AccessModes) == 0 {
+			log.V(1).Info("Cannot set accessMode for new pvc", "namespace", dv.Namespace, "name", dv.Name)
+			recorder.Eventf(dv, v1.EventTypeWarning, ErrClaimNotValid, "DataVolume.storage spec is missing accessMode and no storageClass to choose profile")
+			return nil, errors.Errorf("DataVolume spec is missing accessMode")
+		}
+	} else {
+		pvcSpec.StorageClassName = &storageClass.Name
+		// given storageClass we can apply defaults if needed
+		if (pvcSpec.VolumeMode == nil || *pvcSpec.VolumeMode == "") && (len(pvcSpec.AccessModes) == 0) {
+			accessModes, volumeMode, err := getDefaultVolumeAndAccessMode(client, storageClass)
+			if err != nil {
+				log.V(1).Info("Cannot set accessMode and volumeMode for new pvc", "namespace", dv.Namespace, "name", dv.Name, "Error", err)
+				recorder.Eventf(dv, v1.EventTypeWarning, ErrClaimNotValid,
+					fmt.Sprintf("DataVolume.storage spec is missing accessMode and volumeMode, cannot get access mode from StorageProfile %s", getName(storageClass)))
+				return nil, err
+			}
+			pvcSpec.AccessModes = append(pvcSpec.AccessModes, accessModes...)
+			pvcSpec.VolumeMode = volumeMode
+		} else if len(pvcSpec.AccessModes) == 0 {
+			accessModes, err := getDefaultAccessModes(client, storageClass, pvcSpec.VolumeMode)
+			if err != nil {
+				log.V(1).Info("Cannot set accessMode for new pvc", "namespace", dv.Namespace, "name", dv.Name, "Error", err)
+				recorder.Eventf(dv, v1.EventTypeWarning, ErrClaimNotValid,
+					fmt.Sprintf("DataVolume.storage spec is missing accessMode and cannot get access mode from StorageProfile %s", getName(storageClass)))
+				return nil, err
+			}
+			pvcSpec.AccessModes = append(pvcSpec.AccessModes, accessModes...)
+		} else if pvcSpec.VolumeMode == nil || *pvcSpec.VolumeMode == "" {
+			volumeMode, err := getDefaultVolumeMode(client, storageClass, pvcSpec.AccessModes)
+			if err != nil {
+				return nil, err
+			}
+			pvcSpec.VolumeMode = volumeMode
+		}
+	}
+
+	requestedVolumeSize, err := volumeSize(client, storage, pvcSpec.VolumeMode)
+	if err != nil {
+		return nil, err
+	}
+	if pvcSpec.Resources.Requests == nil {
+		pvcSpec.Resources.Requests = v1.ResourceList{}
+	}
+	pvcSpec.Resources.Requests[v1.ResourceStorage] = *requestedVolumeSize
+
+	return pvcSpec, nil
+}
+
+func copyStorageAsPvc(log logr.Logger, storage *cdiv1.StorageSpec) *v1.PersistentVolumeClaimSpec {
+	input := storage.DeepCopy()
+	pvcSpec := &v1.PersistentVolumeClaimSpec{
+		AccessModes:      input.AccessModes,
+		Selector:         input.Selector,
+		Resources:        input.Resources,
+		VolumeName:       input.VolumeName,
+		StorageClassName: input.StorageClassName,
+		VolumeMode:       input.VolumeMode,
+		DataSource:       input.DataSource,
+	}
+
+	return pvcSpec
+}
+
+func getDefaultVolumeAndAccessMode(c client.Client, storageClass *storagev1.StorageClass) ([]v1.PersistentVolumeAccessMode, *v1.PersistentVolumeMode, error) {
+	if storageClass == nil {
+		return nil, nil, errors.Errorf("no accessMode defined on DV and no StorageProfile")
+	}
+
+	storageProfile := &cdiv1.StorageProfile{}
+	err := c.Get(context.TODO(), types.NamespacedName{Name: storageClass.Name}, storageProfile)
+	if err != nil {
+		return nil, nil, errors.Wrap(err, "cannot get StorageProfile")
+	}
+
+	if len(storageProfile.Status.ClaimPropertySets) > 0 &&
+		len(storageProfile.Status.ClaimPropertySets[0].AccessModes) > 0 {
+		accessModes := storageProfile.Status.ClaimPropertySets[0].AccessModes
+		volumeMode := storageProfile.Status.ClaimPropertySets[0].VolumeMode
+		return accessModes, volumeMode, nil
+	}
+
+	// no accessMode configured on storageProfile
+	return nil, nil, errors.Errorf("no accessMode defined DV nor on StorageProfile for %s StorageClass", storageClass.Name)
+}
+
+func getDefaultVolumeMode(c client.Client, storageClass *storagev1.StorageClass, pvcAccessModes []v1.PersistentVolumeAccessMode) (*v1.PersistentVolumeMode, error) {
+	if storageClass == nil {
+		// fallback to k8s defaults
+		return nil, nil
+	}
+
+	storageProfile := &cdiv1.StorageProfile{}
+	err := c.Get(context.TODO(), types.NamespacedName{Name: storageClass.Name}, storageProfile)
+	if err != nil {
+		return nil, errors.Wrap(err, "cannot get StorageProfile")
+	}
+	if len(storageProfile.Status.ClaimPropertySets) > 0 {
+		volumeMode := storageProfile.Status.ClaimPropertySets[0].VolumeMode
+		if len(pvcAccessModes) == 0 {
+			return volumeMode, nil
+		}
+		// check for volume mode matching with given pvc access modes
+		for _, cps := range storageProfile.Status.ClaimPropertySets {
+			for _, accessMode := range cps.AccessModes {
+				for _, pvcAccessMode := range pvcAccessModes {
+					if accessMode == pvcAccessMode {
+						return cps.VolumeMode, nil
+					}
+				}
+			}
+		}
+		// if not found return default volume mode for the storage class
+		return volumeMode, nil
+	}
+
+	// since volumeMode is optional - > gracefully fallback to k8s defaults,
+	return nil, nil
+}
+
+func getDefaultAccessModes(c client.Client, storageClass *storagev1.StorageClass, pvcVolumeMode *v1.PersistentVolumeMode) ([]v1.PersistentVolumeAccessMode, error) {
+	if storageClass == nil {
+		return nil, errors.Errorf("no accessMode defined on DV, no StorageProfile ")
+	}
+
+	storageProfile := &cdiv1.StorageProfile{}
+	err := c.Get(context.TODO(), types.NamespacedName{Name: storageClass.Name}, storageProfile)
+	if err != nil {
+		return nil, errors.Wrap(err, "no accessMode defined on DV, cannot get StorageProfile")
+	}
+
+	if len(storageProfile.Status.ClaimPropertySets) > 0 {
+		// check for access modes matching with given pvc volume mode
+		defaultAccessModes := []v1.PersistentVolumeAccessMode{}
+		for _, cps := range storageProfile.Status.ClaimPropertySets {
+			if cps.VolumeMode != nil && pvcVolumeMode != nil && *cps.VolumeMode == *pvcVolumeMode {
+				if len(cps.AccessModes) > 0 {
+					return cps.AccessModes, nil
+				}
+			} else if len(cps.AccessModes) > 0 && len(defaultAccessModes) == 0 {
+				defaultAccessModes = cps.AccessModes
+			}
+		}
+		// if not found return default access modes for the storage profile
+		if len(defaultAccessModes) > 0 {
+			return defaultAccessModes, nil
+		}
+	}
+
+	// no accessMode configured on storageProfile
+	return nil, errors.Errorf("no accessMode defined on StorageProfile for %s StorageClass", storageClass.Name)
+}
+
+func volumeSize(c client.Client, storage *cdiv1.StorageSpec, volumeMode *v1.PersistentVolumeMode) (*resource.Quantity, error) {
+	// resources.requests[storage] - just copy it to pvc,
+	requestedSize, found := storage.Resources.Requests[v1.ResourceStorage]
+	if !found {
+		return nil, errors.Errorf("Datavolume Spec is not valid - missing storage size")
+	}
+
+	// disk or image size, inflate it with overhead
+	if resolveVolumeMode(volumeMode) == v1.PersistentVolumeFilesystem {
+		fsOverhead, err := GetFilesystemOverheadForStorageClass(c, storage.StorageClassName)
+		if err != nil {
+			return nil, err
+		}
+		fsOverheadFloat, _ := strconv.ParseFloat(string(fsOverhead), 64)
+		requiredSpace := GetRequiredSpace(fsOverheadFloat, requestedSize.Value())
+
+		return resource.NewScaledQuantity(requiredSpace, 0), nil
+	}
+
+	return &requestedSize, nil
+}
