@@ -1,10 +1,8 @@
 package util
 
 import (
-	"crypto/md5"
+	"bytes"
 	"encoding/base64"
-	"encoding/hex"
-	"io"
 	"io/ioutil"
 	"os"
 	"path/filepath"
@@ -136,9 +134,9 @@ var _ = Describe("Copy files", func() {
 	It("Should copy file from source to dest, with valid source and dest", func() {
 		err = CopyFile(filepath.Join(TestImagesDir, "content.tar"), filepath.Join(destTmp, "target.tar"))
 		Expect(err).ToNot(HaveOccurred())
-		sourceMd5, err := md5sum(filepath.Join(TestImagesDir, "content.tar"))
+		sourceMd5, err := Md5sum(filepath.Join(TestImagesDir, "content.tar"))
 		Expect(err).ToNot(HaveOccurred())
-		targetMd5, err := md5sum(filepath.Join(destTmp, "target.tar"))
+		targetMd5, err := Md5sum(filepath.Join(destTmp, "target.tar"))
 		Expect(err).ToNot(HaveOccurred())
 		Expect(sourceMd5).Should(Equal(targetMd5))
 	})
@@ -154,27 +152,79 @@ var _ = Describe("Copy files", func() {
 	})
 })
 
-func md5sum(filePath string) (string, error) {
-	var returnMD5String string
+var _ = Describe("Zero out ranges in files", func() {
+	var testFile *os.File
+	var testData []byte
+	testData = append(testData, bytes.Repeat([]byte{0x55}, 1024)...)
+	testData = append(testData, bytes.Repeat([]byte{0xAA}, 1024)...)
+	testData = append(testData, bytes.Repeat([]byte{0xFF}, 1024)...)
 
-	file, err := os.Open(filePath)
-	if err != nil {
-		return returnMD5String, err
-	}
-	defer file.Close()
+	BeforeEach(func() {
+		var err error
 
-	hash := md5.New()
+		testFile, err = ioutil.TempFile("", "test")
+		Expect(err).ToNot(HaveOccurred())
+		written, err := testFile.Write(testData)
+		Expect(err).ToNot(HaveOccurred())
+		Expect(written).To(Equal(len(testData)))
+	})
 
-	if _, err := io.Copy(hash, file); err != nil {
-		return returnMD5String, err
-	}
+	AfterEach(func() {
+		testFile.Close()
+		os.Remove(testFile.Name())
+	})
 
-	hashInBytes := hash.Sum(nil)[:16]
-	returnMD5String = hex.EncodeToString(hashInBytes)
+	It("Should successfully zero a range with fallocate", func() {
+		start := 512
+		length := 100
+		end := start + length
+		err := PunchHole(testFile, int64(start), int64(length))
+		Expect(err).ToNot(HaveOccurred())
+		err = testFile.Sync()
+		Expect(err).ToNot(HaveOccurred())
+		err = testFile.Close()
+		Expect(err).ToNot(HaveOccurred())
+		data, err := ioutil.ReadFile(testFile.Name())
+		Expect(err).ToNot(HaveOccurred())
+		Expect(len(data)).To(Equal(len(testData)))
+		comparison := bytes.Compare(data[start:end], bytes.Repeat([]byte{0}, length))
+		Expect(comparison).To(Equal(0))
+		comparison = bytes.Compare(data[0:start], testData[0:start])
+		Expect(comparison).To(Equal(0))
+		comparison = bytes.Compare(data[end:], testData[end:])
+		Expect(comparison).To(Equal(0))
+	})
 
-	return returnMD5String, nil
-}
+	table.DescribeTable("Should successfully append zeroes to a file", func(appendFunction func(f *os.File, start, length int64) error) {
+		length := 1024
+		err := appendFunction(testFile, int64(len(testData)), int64(length))
+		Expect(err).ToNot(HaveOccurred())
+		err = testFile.Sync()
+		Expect(err).ToNot(HaveOccurred())
+		err = testFile.Close()
+		Expect(err).ToNot(HaveOccurred())
+		data, err := ioutil.ReadFile(testFile.Name())
+		Expect(err).ToNot(HaveOccurred())
+		Expect(len(data)).To(Equal(len(testData) + length))
+		comparison := bytes.Compare(data[:len(testData)], testData)
+		Expect(comparison).To(Equal(0))
+		comparison = bytes.Compare(data[len(testData):], bytes.Repeat([]byte{0}, length))
+		Expect(comparison).To(Equal(0))
+	},
+		table.Entry("using truncate", AppendZeroWithTruncate),
+		table.Entry("using write", AppendZeroWithWrite),
+	)
 
+	table.DescribeTable("Should fail to append zeroes to a file using seek if it already has data at the specified starting index", func(appendFunction func(f *os.File, start, length int64) error) {
+		length := 1024
+		err := appendFunction(testFile, 0, int64(length))
+		Expect(err).To(HaveOccurred())
+		Expect(err.Error()).Should(MatchRegexp(".*cannot safely append.*"))
+	},
+		table.Entry("using truncate", AppendZeroWithTruncate),
+		table.Entry("using write", AppendZeroWithWrite),
+	)
+})
 var _ = Describe("Usable Space calculation", func() {
 
 	const (
