@@ -4,12 +4,15 @@ import (
 	"context"
 	"encoding/json"
 	"fmt"
+	"io/ioutil"
 	"reflect"
+	"strconv"
 	"time"
 
 	. "github.com/onsi/ginkgo"
 	. "github.com/onsi/gomega"
 
+	promv1 "github.com/coreos/prometheus-operator/pkg/apis/monitoring/v1"
 	routev1 "github.com/openshift/api/route/v1"
 	routeclient "github.com/openshift/client-go/route/clientset/versioned"
 	secclient "github.com/openshift/client-go/security/clientset/versioned"
@@ -17,13 +20,19 @@ import (
 	appsv1 "k8s.io/api/apps/v1"
 	corev1 "k8s.io/api/core/v1"
 	rbacv1 "k8s.io/api/rbac/v1"
+	schedulev1 "k8s.io/api/scheduling/v1"
+	storagev1 "k8s.io/api/storage/v1"
 	"k8s.io/apimachinery/pkg/api/errors"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/labels"
+	"k8s.io/apimachinery/pkg/types"
+	crclient "sigs.k8s.io/controller-runtime/pkg/client"
 
-	cdiv1 "kubevirt.io/containerized-data-importer/pkg/apis/core/v1beta1"
+	cdiv1 "kubevirt.io/containerized-data-importer-api/pkg/apis/core/v1beta1"
 	cdiClientset "kubevirt.io/containerized-data-importer/pkg/client/clientset/versioned"
 	"kubevirt.io/containerized-data-importer/pkg/common"
+	"kubevirt.io/containerized-data-importer/pkg/controller"
+	resourcesutils "kubevirt.io/containerized-data-importer/pkg/operator/resources/utils"
 	"kubevirt.io/containerized-data-importer/tests"
 	"kubevirt.io/containerized-data-importer/tests/framework"
 	"kubevirt.io/containerized-data-importer/tests/utils"
@@ -54,7 +63,7 @@ var _ = Describe("ALL Operator tests", func() {
 				promService, err := f.K8sClient.CoreV1().Services(f.CdiInstallNs).Get(context.TODO(), common.PrometheusServiceName, metav1.GetOptions{})
 				Expect(err).ToNot(HaveOccurred())
 				Expect(promService.Spec.Ports[0].Name).To(Equal("metrics"))
-				Expect(promService.Spec.Selector[common.PrometheusLabel]).To(Equal(""))
+				Expect(promService.Spec.Selector[common.PrometheusLabelKey]).To(Equal(common.PrometheusLabelValue))
 				originalTimeStamp := promService.ObjectMeta.CreationTimestamp
 
 				By("Deleting the service")
@@ -69,7 +78,7 @@ var _ = Describe("ALL Operator tests", func() {
 					return false
 				}, 1*time.Minute, 2*time.Second).Should(BeTrue())
 				Expect(promService.Spec.Ports[0].Name).To(Equal("metrics"))
-				Expect(promService.Spec.Selector[common.PrometheusLabel]).To(Equal(""))
+				Expect(promService.Spec.Selector[common.PrometheusLabelKey]).To(Equal(common.PrometheusLabelValue))
 			})
 
 			It("[test_id:3952]add cdi-sa to containerized-data-importer scc", func() {
@@ -87,7 +96,7 @@ var _ = Describe("ALL Operator tests", func() {
 				Expect(scc.Users).Should(ContainElement(cdiSA))
 			})
 
-			// Condition flags can be found here with their meaning https://github.com/kubevirt/hyperconverged-cluster-operator/blob/master/docs/conditions.md
+			// Condition flags can be found here with their meaning https://github.com/kubevirt/hyperconverged-cluster-operator/blob/main/docs/conditions.md
 			It("[test_id:3953]Condition flags on CR should be healthy and operating", func() {
 				cdiObjects, err := f.CdiClient.CdiV1beta1().CDIs().List(context.TODO(), metav1.ListOptions{})
 				Expect(err).ToNot(HaveOccurred())
@@ -286,15 +295,18 @@ var _ = Describe("ALL Operator tests", func() {
 					return
 				}
 
+				By("Check if CDI CR exists")
 				cdi, err := f.CdiClient.CdiV1beta1().CDIs().Get(context.TODO(), cr.Name, metav1.GetOptions{})
 				if err == nil {
 					if cdi.DeletionTimestamp == nil {
+						By("CDI CR exists")
 						cdi.Spec = cr.Spec
 						_, err = f.CdiClient.CdiV1beta1().CDIs().Update(context.TODO(), cdi, metav1.UpdateOptions{})
 						Expect(err).ToNot(HaveOccurred())
 						return
 					}
 
+					By("Waiting for CDI CR deletion")
 					Eventually(func() bool {
 						_, err = f.CdiClient.CdiV1beta1().CDIs().Get(context.TODO(), cr.Name, metav1.GetOptions{})
 						if errors.IsNotFound(err) {
@@ -314,9 +326,11 @@ var _ = Describe("ALL Operator tests", func() {
 					Spec: cr.Spec,
 				}
 
+				By("Create CDI CR")
 				cdi, err = f.CdiClient.CdiV1beta1().CDIs().Create(context.TODO(), cdi, metav1.CreateOptions{})
 				Expect(err).ToNot(HaveOccurred())
 
+				By("Waiting for CDI CR")
 				Eventually(func() bool {
 					cdi, err = f.CdiClient.CdiV1beta1().CDIs().Get(context.TODO(), cr.Name, metav1.GetOptions{})
 					Expect(err).ToNot(HaveOccurred())
@@ -409,18 +423,85 @@ var _ = Describe("ALL Operator tests", func() {
 				Expect(err).ToNot(HaveOccurred())
 				f.ForceBindPvcIfDvIsWaitForFirstConsumer(dv)
 
+				By("Creating datavolume with DataImportCron label")
+				dv = utils.NewDataVolumeForUpload("retain-me", "1Gi")
+				dv.Labels = map[string]string{common.DataImportCronLabel: "dic"}
+				dv, err = utils.CreateDataVolumeFromDefinition(f.CdiClient, f.Namespace.Name, dv)
+				Expect(err).ToNot(HaveOccurred())
+				f.ForceBindPvcIfDvIsWaitForFirstConsumer(dv)
+
 				By("Cannot delete CDI")
 				err = f.CdiClient.CdiV1beta1().CDIs().Delete(context.TODO(), cr.Name, metav1.DeleteOptions{DryRun: []string{"All"}})
 				Expect(err).To(HaveOccurred())
 				Expect(err.Error()).To(ContainSubstring("there are still DataVolumes present"))
 
-				err = f.CdiClient.CdiV1beta1().DataVolumes(f.Namespace.Name).Delete(context.TODO(), dv.Name, metav1.DeleteOptions{})
+				By("Delete the unlabeled datavolume")
+				err = f.CdiClient.CdiV1beta1().DataVolumes(f.Namespace.Name).Delete(context.TODO(), "delete-me", metav1.DeleteOptions{})
 				Expect(err).ToNot(HaveOccurred())
 
 				By("Can delete CDI")
 				err = f.CdiClient.CdiV1beta1().CDIs().Delete(context.TODO(), cr.Name, metav1.DeleteOptions{DryRun: []string{"All"}})
 				Expect(err).ToNot(HaveOccurred())
 			})
+
+			It("[test_id:8087]CDI CR deletion should delete DataImportCron CRD and all DataImportCrons", func() {
+				reg, err := getDataVolumeSourceRegistry(f)
+				Expect(err).To(BeNil())
+
+				By("Create new DataImportCron")
+				cron := NewDataImportCron("cron-test", "5Gi", scheduleEveryMinute, "ds", *reg)
+				cron, err = f.CdiClient.CdiV1beta1().DataImportCrons(f.Namespace.Name).Create(context.TODO(), cron, metav1.CreateOptions{})
+				Expect(err).ToNot(HaveOccurred())
+
+				By("Verify cron first import completed")
+				Eventually(func() bool {
+					cron, err = f.CdiClient.CdiV1beta1().DataImportCrons(f.Namespace.Name).Get(context.TODO(), cron.Name, metav1.GetOptions{})
+					Expect(err).ToNot(HaveOccurred())
+					upToDateCond := controller.FindDataImportCronConditionByType(cron, cdiv1.DataImportCronUpToDate)
+					return upToDateCond != nil && upToDateCond.Status == corev1.ConditionTrue
+				}, dataImportCronTimeout, pollingInterval).Should(BeTrue())
+
+				pvc := cron.Status.LastImportedPVC
+				Expect(pvc).ToNot(BeNil())
+
+				By("Verify dv succeeded")
+				dv, err := f.CdiClient.CdiV1beta1().DataVolumes(pvc.Namespace).Get(context.TODO(), pvc.Name, metav1.GetOptions{})
+				Expect(err).ToNot(HaveOccurred())
+				Expect(dv.Status.Phase).To(Equal(cdiv1.Succeeded))
+
+				By("Start goroutine creating DataImportCrons")
+				go func() {
+					defer GinkgoRecover()
+					var err error
+					for i := 0; i < 100 && err == nil; i++ {
+						cronName := fmt.Sprintf("cron-test-%d", i)
+						cron := NewDataImportCron(cronName, "5Gi", scheduleEveryMinute, "ds", *reg)
+						_, err = f.CdiClient.CdiV1beta1().DataImportCrons(f.Namespace.Name).Create(context.TODO(), cron, metav1.CreateOptions{})
+					}
+				}()
+
+				removeCDI()
+
+				By("Verify no DataImportCrons are found")
+				Eventually(func() bool {
+					_, err = f.CdiClient.CdiV1beta1().DataImportCrons(f.Namespace.Name).List(context.TODO(), metav1.ListOptions{})
+					return err != nil && errors.IsNotFound(err)
+				}, 1*time.Minute, 2*time.Second).Should(BeTrue())
+
+				By("Verify no cronjobs left")
+				Eventually(func() bool {
+					cronjobs, err := f.K8sClient.BatchV1().CronJobs(f.CdiInstallNs).List(context.TODO(), metav1.ListOptions{})
+					Expect(err).ToNot(HaveOccurred())
+					return len(cronjobs.Items) == 0
+				}, 1*time.Minute, 2*time.Second).Should(BeTrue())
+
+				By("Verify no jobs left")
+				Eventually(func() bool {
+					jobs, err := f.K8sClient.BatchV1().Jobs(f.CdiInstallNs).List(context.TODO(), metav1.ListOptions{})
+					Expect(err).ToNot(HaveOccurred())
+					return len(jobs.Items) == 0
+				}, 1*time.Minute, 2*time.Second).Should(BeTrue())
+			})
 		})
 
 		var _ = Describe("[rfe_id:4784][crit:high] CDI Operator deployment + CDI CR delete tests", func() {
@@ -667,6 +748,247 @@ var _ = Describe("ALL Operator tests", func() {
 			})
 		})
 
+		var _ = Describe("[rfe_id:7101][crit:medium][vendor:cnv-qe@redhat.com][level:component]Alert tests", func() {
+			var numAddedStorageClasses int
+			f := framework.NewFramework("alert-tests")
+
+			AfterEach(func() {
+				By("Delete unknown storage classes")
+				for i := 0; i < numAddedStorageClasses; i++ {
+					name := fmt.Sprintf("unknown-sc-%d", i)
+					_, err := f.K8sClient.StorageV1().StorageClasses().Get(context.TODO(), name, metav1.GetOptions{})
+					if err != nil && errors.IsNotFound(err) {
+						continue
+					}
+					err = f.K8sClient.StorageV1().StorageClasses().Delete(context.TODO(), name, metav1.DeleteOptions{})
+					Expect(err).ToNot(HaveOccurred())
+				}
+			})
+
+			getMetricValue := func(endpoint string) int {
+				var returnVal string
+
+				Eventually(func() bool {
+					var result map[string]interface{}
+					resp := tests.MakePrometheusHTTPRequest(f, "query?query="+endpoint)
+					defer resp.Body.Close()
+					bodyBytes, err := ioutil.ReadAll(resp.Body)
+					if err != nil {
+						return false
+					}
+					err = json.Unmarshal(bodyBytes, &result)
+					if err != nil {
+						return false
+					}
+					if len(result["data"].(map[string]interface{})["result"].([]interface{})) == 0 {
+						return false
+					}
+					values := result["data"].(map[string]interface{})["result"].([]interface{})[0].(map[string]interface{})["value"].([]interface{})
+					for _, v := range values {
+						if s, ok := v.(string); ok {
+							returnVal = s
+							return true
+						}
+					}
+					return false
+				}, 1*time.Minute, 1*time.Second).Should(BeTrue())
+
+				i, err := strconv.Atoi(returnVal)
+				Expect(err).ToNot(HaveOccurred())
+				return i
+			}
+
+			createUnknownStorageClass := func(name string) *storagev1.StorageClass {
+				immediateBinding := storagev1.VolumeBindingImmediate
+
+				return &storagev1.StorageClass{
+					ObjectMeta: metav1.ObjectMeta{
+						Name: name,
+						Labels: map[string]string{
+							"cdi.kubevirt.io/testing": "",
+						},
+					},
+					Provisioner:       "kubernetes.io/non-existent-provisioner",
+					VolumeBindingMode: &immediateBinding,
+				}
+			}
+
+			It("[test_id:7962] CDIOperatorDown alert firing when operator scaled down", func() {
+				if !tests.IsPrometheusAvailable(f.ExtClient) {
+					Skip("This test depends on prometheus infra being available")
+				}
+
+				By("Scale down operator so alert will trigger")
+				deploymentName := "cdi-operator"
+				operatorDeployment, err := f.K8sClient.AppsV1().Deployments(f.CdiInstallNs).Get(context.TODO(), deploymentName, metav1.GetOptions{})
+				Expect(err).ToNot(HaveOccurred())
+				originalReplicas := operatorDeployment.Spec.Replicas
+				operatorDeployment.Spec.Replicas = &[]int32{0}[0]
+				_, err = f.K8sClient.AppsV1().Deployments(f.CdiInstallNs).Update(context.TODO(), operatorDeployment, metav1.UpdateOptions{})
+				Expect(err).ToNot(HaveOccurred())
+				Eventually(func() bool {
+					dep, err := f.K8sClient.AppsV1().Deployments(f.CdiInstallNs).Get(context.TODO(), deploymentName, metav1.GetOptions{})
+					Expect(err).ToNot(HaveOccurred())
+					return dep.Status.Replicas == 0
+				}, 20*time.Second, 1*time.Second).Should(BeTrue())
+				By("Patch our rule so alert fires a little faster")
+				promRule := &promv1.PrometheusRule{
+					ObjectMeta: metav1.ObjectMeta{
+						Name:      "prometheus-cdi-rules",
+						Namespace: f.CdiInstallNs,
+					},
+				}
+				err = f.CrClient.Get(context.TODO(), crclient.ObjectKeyFromObject(promRule), promRule)
+				Expect(err).ToNot(HaveOccurred())
+				for i, group := range promRule.Spec.Groups {
+					if group.Name == "cdi.rules" {
+						for j, rule := range group.Rules {
+							if rule.Alert == "CDIOperatorDown" {
+								rule.For = "1m"
+								promRule.Spec.Groups[i].Rules[j] = rule
+								break
+							}
+						}
+						break
+					}
+				}
+				err = f.CrClient.Update(context.TODO(), promRule)
+				Expect(err).ToNot(HaveOccurred())
+
+				By("Let's see that alert fires")
+				Eventually(func() bool {
+					var result map[string]interface{}
+					resp := tests.MakePrometheusHTTPRequest(f, "alerts")
+					defer resp.Body.Close()
+					// Make sure alert appears and is firing
+					bodyBytes, err := ioutil.ReadAll(resp.Body)
+					if err != nil {
+						return false
+					}
+					err = json.Unmarshal(bodyBytes, &result)
+					if err != nil {
+						return false
+					}
+					alerts := result["data"].(map[string]interface{})["alerts"].([]interface{})
+					for _, alert := range alerts {
+						name := alert.(map[string]interface{})["labels"].(map[string]interface{})["alertname"].(string)
+						if name == "CDIOperatorDown" {
+							if state := alert.(map[string]interface{})["state"].(string); state == "firing" {
+								return true
+							}
+						}
+					}
+					return false
+				}, 10*time.Minute, 1*time.Second).Should(BeTrue())
+
+				By("Ensuring original value of replicas restored")
+				operatorDeployment, err = f.K8sClient.AppsV1().Deployments(f.CdiInstallNs).Get(context.TODO(), deploymentName, metav1.GetOptions{})
+				Expect(err).ToNot(HaveOccurred())
+				operatorDeployment.Spec.Replicas = originalReplicas
+				_, err = f.K8sClient.AppsV1().Deployments(f.CdiInstallNs).Update(context.TODO(), operatorDeployment, metav1.UpdateOptions{})
+				Expect(err).ToNot(HaveOccurred())
+				err = utils.WaitForDeploymentReplicasReady(f.K8sClient, f.CdiInstallNs, deploymentName)
+				Expect(err).ToNot(HaveOccurred())
+			})
+
+			It("[test_id:7963] CDI ready metric value as expected when ready to use", func() {
+				if !tests.IsPrometheusAvailable(f.ExtClient) {
+					Skip("This test depends on prometheus infra being available")
+				}
+
+				Eventually(func() int {
+					return getMetricValue("kubevirt_cdi_cr_ready")
+				}, 1*time.Minute, 1*time.Second).Should(BeNumerically("==", 1))
+			})
+
+			It("[test_id:7965] StorageProfile incomplete metric expected value when creating an incomplete profile", func() {
+				if !tests.IsPrometheusAvailable(f.ExtClient) {
+					Skip("This test depends on prometheus infra being available")
+				}
+
+				numAddedStorageClasses = 2
+				defaultStorageClass := utils.GetDefaultStorageClass(f.K8sClient)
+				defaultStorageClassProfile := &cdiv1.StorageProfile{}
+				err := f.CrClient.Get(context.TODO(), types.NamespacedName{Name: defaultStorageClass.Name}, defaultStorageClassProfile)
+				Expect(err).ToNot(HaveOccurred())
+				for i := 0; i < numAddedStorageClasses; i++ {
+					_, err = f.K8sClient.StorageV1().StorageClasses().Create(context.TODO(), createUnknownStorageClass(fmt.Sprintf("unknown-sc-%d", i)), metav1.CreateOptions{})
+					Expect(err).ToNot(HaveOccurred())
+				}
+				originalMetricVal := getMetricValue("kubevirt_cdi_incomplete_storageprofiles_total")
+				expectedIncomplete := originalMetricVal + numAddedStorageClasses
+
+				Eventually(func() int {
+					return getMetricValue("kubevirt_cdi_incomplete_storageprofiles_total")
+				}, 2*time.Minute, 1*time.Second).Should(BeNumerically("==", expectedIncomplete))
+
+				By("Fix profiles to be complete and test metric value equals original")
+				for i := 0; i < numAddedStorageClasses; i++ {
+					profile := &cdiv1.StorageProfile{}
+					err = f.CrClient.Get(context.TODO(), types.NamespacedName{Name: fmt.Sprintf("unknown-sc-%d", i)}, profile)
+					Expect(err).ToNot(HaveOccurred())
+					// These might be wrong values, but at least the profile is no longer incomplete
+					profile.Spec.ClaimPropertySets = defaultStorageClassProfile.Status.ClaimPropertySets
+					err = f.CrClient.Update(context.TODO(), profile)
+					Expect(err).ToNot(HaveOccurred())
+					expectedIncomplete--
+					Eventually(func() int {
+						return getMetricValue("kubevirt_cdi_incomplete_storageprofiles_total")
+					}, 2*time.Minute, 1*time.Second).Should(BeNumerically("==", expectedIncomplete))
+				}
+			})
+
+			It("[test_id:7964] DataImportCron failing metric expected value when patching DesiredDigest annotation with junk sha256 value", func() {
+				if !tests.IsPrometheusAvailable(f.ExtClient) {
+					Skip("This test depends on prometheus infra being available")
+				}
+				numCrons := 2
+				originalMetricVal := getMetricValue("kubevirt_cdi_dataimportcron_outdated_total")
+				expectedFailingCrons := originalMetricVal + numCrons
+
+				reg, err := getDataVolumeSourceRegistry(f)
+				Expect(err).To(BeNil())
+				defer utils.RemoveInsecureRegistry(f.CrClient, *reg.URL)
+
+				for i := 1; i < numCrons+1; i++ {
+					cron := NewDataImportCron(fmt.Sprintf("cron-test-%d", i), "5Gi", scheduleOnceAYear, fmt.Sprintf("datasource-test-%d", i), *reg)
+					By(fmt.Sprintf("Create new DataImportCron %s", *reg.URL))
+					cron, err = f.CdiClient.CdiV1beta1().DataImportCrons(f.Namespace.Name).Create(context.TODO(), cron, metav1.CreateOptions{})
+					Expect(err).ToNot(HaveOccurred())
+					By("Wait for condition UpToDate=true on DataImportCron")
+					Eventually(func() bool {
+						cron, err = f.CdiClient.CdiV1beta1().DataImportCrons(f.Namespace.Name).Get(context.TODO(), cron.Name, metav1.GetOptions{})
+						Expect(err).ToNot(HaveOccurred())
+						upToDateCondition := controller.FindDataImportCronConditionByType(cron, cdiv1.DataImportCronUpToDate)
+						return upToDateCondition != nil && upToDateCondition.ConditionState.Status == corev1.ConditionTrue
+					}, dataImportCronTimeout, pollingInterval).Should(BeTrue())
+					Eventually(func() error {
+						cron, err = f.CdiClient.CdiV1beta1().DataImportCrons(f.Namespace.Name).Get(context.TODO(), cron.Name, metav1.GetOptions{})
+						Expect(err).ToNot(HaveOccurred())
+						if cron.Annotations == nil {
+							cron.Annotations = make(map[string]string)
+						}
+						// notarealsha
+						cron.Annotations[controller.AnnSourceDesiredDigest] = "sha256:c616e1c85568b1f1d528da2b3dbc257fd2035ada441e286a8c42606491442a5d"
+						cron, err = f.CdiClient.CdiV1beta1().DataImportCrons(f.Namespace.Name).Update(context.TODO(), cron, metav1.UpdateOptions{})
+						return err
+					}, dataImportCronTimeout, pollingInterval).Should(BeNil())
+					By(fmt.Sprintf("Ensuring metric value incremented to %d", originalMetricVal+i))
+					Eventually(func() int {
+						return getMetricValue("kubevirt_cdi_dataimportcron_outdated_total")
+					}, 3*time.Minute, 1*time.Second).Should(BeNumerically("==", originalMetricVal+i))
+				}
+				By("Ensure metric value decrements when crons are cleaned up")
+				for i := 1; i < numCrons+1; i++ {
+					err = f.CdiClient.CdiV1beta1().DataImportCrons(f.Namespace.Name).Delete(context.TODO(), fmt.Sprintf("cron-test-%d", i), metav1.DeleteOptions{})
+					Expect(err).ToNot(HaveOccurred())
+					Eventually(func() int {
+						return getMetricValue("kubevirt_cdi_dataimportcron_outdated_total")
+					}, 3*time.Minute, 1*time.Second).Should(BeNumerically("==", expectedFailingCrons-i))
+				}
+			})
+		})
+
 		var _ = Describe("Operator cert config tests", func() {
 			var cdi *cdiv1.CDI
 			f := framework.NewFramework("operator-cert-config-test")
@@ -802,6 +1124,119 @@ var _ = Describe("ALL Operator tests", func() {
 				}, 2*time.Minute, 1*time.Second).Should(BeTrue())
 			})
 		})
+
+		var _ = Describe("Priority class tests", func() {
+			var (
+				cdi                   *cdiv1.CDI
+				systemClusterCritical = cdiv1.CDIPriorityClass("system-cluster-critical")
+				osUserCrit            = &schedulev1.PriorityClass{
+					ObjectMeta: metav1.ObjectMeta{
+						Name: resourcesutils.CDIPriorityClass,
+					},
+					Value: 10000,
+				}
+			)
+			f := framework.NewFramework("operator-priority-class-test")
+			verifyPodPriorityClass := func(prefix, priorityClassName, labelSelector string) {
+				Eventually(func() string {
+					controllerPod, err := utils.FindPodByPrefix(f.K8sClient, f.CdiInstallNs, prefix, labelSelector)
+					if err != nil {
+						return ""
+					}
+					return controllerPod.Spec.PriorityClassName
+				}, 2*time.Minute, 1*time.Second).Should(BeEquivalentTo(priorityClassName))
+			}
+
+			BeforeEach(func() {
+				cr, err := f.CdiClient.CdiV1beta1().CDIs().Get(context.TODO(), "cdi", metav1.GetOptions{})
+				if errors.IsNotFound(err) {
+					Skip("CDI CR 'cdi' does not exist.  Probably managed by another operator so skipping.")
+				}
+				Expect(err).ToNot(HaveOccurred())
+				cdi = cr
+				if cr.Spec.PriorityClass != nil {
+					By(fmt.Sprintf("Current priority class is: [%s]", *cr.Spec.PriorityClass))
+				}
+			})
+
+			AfterEach(func() {
+				if cdi == nil {
+					return
+				}
+
+				cr, err := f.CdiClient.CdiV1beta1().CDIs().Get(context.TODO(), "cdi", metav1.GetOptions{})
+				Expect(err).ToNot(HaveOccurred())
+
+				cr.Spec.PriorityClass = cdi.Spec.PriorityClass
+
+				_, err = f.CdiClient.CdiV1beta1().CDIs().Update(context.TODO(), cr, metav1.UpdateOptions{})
+				Expect(err).ToNot(HaveOccurred())
+
+				if !utils.IsOpenshift(f.K8sClient) {
+					Eventually(func() bool {
+						return errors.IsNotFound(f.K8sClient.SchedulingV1().PriorityClasses().Delete(context.TODO(), osUserCrit.Name, metav1.DeleteOptions{}))
+					}, 2*time.Minute, 1*time.Second).Should(BeTrue())
+				}
+				By("Ensuring the CDI priority class is restored")
+				prioClass := ""
+				if cr.Spec.PriorityClass != nil {
+					prioClass = string(*cr.Spec.PriorityClass)
+				}
+				// Deployment
+				verifyPodPriorityClass(cdiDeploymentPodPrefix, string(prioClass), common.CDILabelSelector)
+				// API server
+				verifyPodPriorityClass(cdiApiServerPodPrefix, string(prioClass), "")
+				// Upload server
+				verifyPodPriorityClass(cdiUploadProxyPodPrefix, string(prioClass), "")
+				By("Verifying there is just a single cdi controller pod")
+				Eventually(func() error {
+					_, err := utils.FindPodByPrefix(f.K8sClient, f.CdiInstallNs, cdiDeploymentPodPrefix, common.CDILabelSelector)
+					return err
+				}, 2*time.Minute, 1*time.Second).Should(BeNil())
+				By("Ensuring this pod is the leader")
+				Eventually(func() bool {
+					controllerPod, err := utils.FindPodByPrefix(f.K8sClient, f.CdiInstallNs, cdiDeploymentPodPrefix, common.CDILabelSelector)
+					Expect(err).ToNot(HaveOccurred())
+					log := getLog(f, controllerPod.Name)
+					return checkLogForRegEx(logIsLeaderRegex, log)
+				}, 2*time.Minute, 1*time.Second).Should(BeTrue())
+
+			})
+
+			It("should use kubernetes priority class if set", func() {
+				cr, err := f.CdiClient.CdiV1beta1().CDIs().Get(context.TODO(), "cdi", metav1.GetOptions{})
+				if errors.IsNotFound(err) {
+					Skip("CDI CR 'cdi' does not exist.  Probably managed by another operator so skipping.")
+				}
+				Expect(err).ToNot(HaveOccurred())
+				By("Setting the priority class to system cluster critical, which is known to exist")
+				cr.Spec.PriorityClass = &systemClusterCritical
+				_, err = f.CdiClient.CdiV1beta1().CDIs().Update(context.TODO(), cr, metav1.UpdateOptions{})
+				Expect(err).ToNot(HaveOccurred())
+				By("Verifying the CDI deployment is updated")
+				verifyPodPriorityClass(cdiDeploymentPodPrefix, string(systemClusterCritical), common.CDILabelSelector)
+				By("Verifying the CDI api server is updated")
+				verifyPodPriorityClass(cdiApiServerPodPrefix, string(systemClusterCritical), "")
+				By("Verifying the CDI upload proxy server is updated")
+				verifyPodPriorityClass(cdiUploadProxyPodPrefix, string(systemClusterCritical), "")
+			})
+
+			It("should use openshift priority class if not set and available", func() {
+				_, err := f.CdiClient.CdiV1beta1().CDIs().Get(context.TODO(), "cdi", metav1.GetOptions{})
+				if errors.IsNotFound(err) {
+					Skip("CDI CR 'cdi' does not exist.  Probably managed by another operator so skipping.")
+				}
+				_, err = f.K8sClient.SchedulingV1().PriorityClasses().Create(context.TODO(), osUserCrit, metav1.CreateOptions{})
+				Expect(err).ToNot(HaveOccurred())
+				By("Verifying the CDI control plane is updated")
+				// Deployment
+				verifyPodPriorityClass(cdiDeploymentPodPrefix, string(osUserCrit.Name), common.CDILabelSelector)
+				// API server
+				verifyPodPriorityClass(cdiApiServerPodPrefix, string(osUserCrit.Name), "")
+				// Upload server
+				verifyPodPriorityClass(cdiUploadProxyPodPrefix, string(osUserCrit.Name), "")
+			})
+		})
 	})
 })
 
